// The View class for XaoS on BeOS.
// Copyright  1997  Jens Kilian (jjk@acm.org)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <AppDefs.h>
#include <Bitmap.h>
#include <Directory.h>
#include <File.h>
#include <InterfaceDefs.h>
#include <OS.h>
#include <Message.h>
#include <Point.h>
#include <Rect.h>
#include <StorageDefs.h>
#include <SupportDefs.h>
#include <View.h>
#include <Volume.h>

#include "CLocker.h"
#include "XaoSView.h"
#include "XaoSEvent.h"

#include "ui.h"
#include "version.h"

XaoSView::XaoSView(float width, float height)
:	BView(BRect(B_ORIGIN, BPoint(width-1, height-1)),
		  "XaoS " XaoS_VERSION,
		  B_FOLLOW_ALL_SIDES,
		  B_WILL_DRAW|B_PULSE_NEEDED|B_FRAME_EVENTS),
	mEventPort(create_port(B_MAX_PORT_COUNT, "XaoS " XaoS_VERSION)),
	mpBuffer(0)
{
	// empty
}

XaoSView::~XaoSView(void)
{
	// empty
}

void XaoSView::AttachedToWindow(void)
{
	BView::AttachedToWindow();

	// We will always Draw() the whole window.
	SetViewColor(B_TRANSPARENT_32_BIT);
	
	// Our ugly help text will be truly ugly.
	SetFontName("Kate");
	SetFontSize(9);
	GetFontInfo(&mFontInfo);
	
	SetDrawingMode(B_OP_COPY);
	SetLowColor(255, 255, 255);
	SetHighColor(0, 0, 0);
	
	// Arrgh.  See Pulse().
	Window()->SetPulseRate(100.0 * 1000.0);
}

void XaoSView::Draw(BRect updateRect)
{
	if (mpBuffer) {
		DrawBitmap(mpBuffer, updateRect, updateRect);
	}
}

void XaoSView::FrameResized(float /*width*/, float /*height*/)
{
	SendEvent(XaoSEvent::Resize, XaoSEvent());
}

void XaoSView::KeyDown(ulong aChar)
{
	SendEvent(XaoSEvent::Key, XaoSEvent(aChar));
}

void XaoSView::MessageReceived(BMessage *pMessage)
{
	switch (pMessage->what) {
	
	case B_MESSAGE_NOT_UNDERSTOOD:
		{	BMessage *pPrevious = pMessage->Previous();
			if (pPrevious && pPrevious->what == 'PPNT') {
				// Dragged PPNT message not accepted.
				// Retry using the B_SIMPLE_DATA protocol.
				BMessage *pReply = ConvertDraggedMessage(pPrevious);
				if (pReply) {
					pMessage->SendReply(pReply);
				}
			}
		}
		break;

	default:
		BView::MessageReceived(pMessage);
		break;
	}
}

void XaoSView::MouseDown(BPoint point)
{
	long clicks = Window()->CurrentMessage()->FindLong("clicks");
	if (mpBuffer && clicks >= 2) {
		// Initiate dragging of the picture, in ImageViewer fashion.
		// Undocumented, but everybody seems to know about it.
		BMessage *pMessage = new BMessage('PPNT');
		BRect bounds = mpBuffer->Bounds();
		pMessage->AddRect("bounds", bounds);
		pMessage->AddLong("color_space", mpBuffer->ColorSpace());
		pMessage->AddLong("bits_length", mpBuffer->BitsLength());
		pMessage->AddData("bits", B_RAW_TYPE,
						  mpBuffer->Bits(), mpBuffer->BitsLength());
		
		BBitmap *pDragImage = new BBitmap(bounds, B_COLOR_8_BIT, TRUE);
		BView *const pView =
			new BView(bounds, "BBitmap view", B_FOLLOW_NONE, 0);
		pDragImage->AddChild(pView);
		pView->Window()->Lock();
		pView->DrawBitmap(mpBuffer, B_ORIGIN);
		pView->Sync();
		pView->Window()->Unlock();
							
		DragMessage(pMessage, pDragImage, point);

	} else {
		// Initiate mouse tracking.  Be recommends doing it this way,
		// leaving the window locked; we must be able to draw from the
		// main thread, so we temporarily unlock the window.
		// We also must delay sending the first event until after
		// we're sure that this is not the start of a double-click.
		// I'm not entirely happy about this.
		Window()->Unlock();

		const double sample_interval = 50.0 * 1000.0;
		double delay;
		if (get_click_speed(&delay) != B_NO_ERROR) {
			delay = 100.0 * 1000.0;
		}
		ulong buttons = Window()->CurrentMessage()->FindLong("buttons");
		while (buttons) {
			if (delay <= 0) {
				SendEvent(XaoSEvent::Mouse,
						  XaoSEvent((int)point.x, (int)point.y,
									buttons, modifiers()));
			}
			snooze(sample_interval);
			delay -= sample_interval;
			{	CLocker locker(*Window());
				GetMouse(&point, &buttons, TRUE);
			}
		}
		Window()->Lock();

		SendEvent(XaoSEvent::Mouse,
				  XaoSEvent((int)point.x, (int)point.y, 0, 0));
	}
}

void XaoSView::Pulse(void)
{
	// Arrgh.  BeOS DR8.2 doesn't deliver key-up events, so we have to
	// tell the driver to periodically check the cursor keys.
	SendEvent(XaoSEvent::Pulse, XaoSEvent());
}

void XaoSView::WindowActivated(bool active)
{
	BView::WindowActivated(active);
	if (active) {
		MakeFocus();
	}
}

void XaoSView::SetBuffer(BBitmap *pBuffer)
{
	mpBuffer = pBuffer;
}

void XaoSView::DrawText(BPoint leftTop, const char *const text)
{
	float width = StringWidth(text);
	float height = TextHeight();
	FillRect(BRect(leftTop, leftTop + BPoint(width, height)),
				B_SOLID_LOW);
	if (*(long *)text == 'BeOS') {
		// A little surprise :-)
		SetHighColor(0, 0, 255);
		DrawChar('B', leftTop + BPoint(0, mFontInfo.ascent));
		SetHighColor(255, 0, 0);
		DrawChar('e');
		SetHighColor(0, 0, 0);
		DrawString(text+2);

	} else {
		DrawString(text, leftTop + BPoint(0, mFontInfo.ascent));
	}
}

BMessage *XaoSView::ConvertDraggedMessage(BMessage *pOriginal)
{
	// The receiver of a dragged PPNT message didn't understand it.
	// There is another semi-standard protocol for dragging a bitmap:
	// put it in a file and send a B_SIMPLE_DATA message containing
	// the file's reference.  The file should be in screen-dump (BITM)
	// format.
	
	// The following code was snarfed from William Adams' Rraster
	// tutorial (see ftp://ftp.be.com/pub/Samples/).
	BVolume		boot_vol;
	BDirectory	root_dir;
	BDirectory	tmp_dir;

	boot_vol = boot_volume();
	boot_vol.GetRootDirectory(&root_dir);
	if (root_dir.GetDirectory("tmp", &tmp_dir) != B_NO_ERROR) {
		root_dir.Create("tmp", &tmp_dir);
	}
	if (root_dir.Error() == B_NO_ERROR) {
		const char filename[] = "Fractal";
		
		BFile file;
		long error = tmp_dir.Create(filename, &file);
		if (error != B_NO_ERROR) {
			error = tmp_dir.GetFile(filename, &file);
		}
		if (error == B_NO_ERROR) {
			error = file.Open(B_READ_WRITE);
		}
		if (error == B_NO_ERROR) {
			// Wow, we actually managed to open the file.
			// Now to write some stuff into it.
			struct {
				long left, top, right, bottom;
				long color_space;
				long bits_length;
			} BITMheader;
			
			BRect bounds = pOriginal->FindRect("bounds");
			BITMheader.left = bounds.left;
			BITMheader.top = bounds.top;
			BITMheader.right = bounds.right;
			BITMheader.bottom = bounds.bottom;
			
			BITMheader.color_space = pOriginal->FindLong("color_space");
			BITMheader.bits_length = pOriginal->FindLong("bits_length");
			
			file.Write(&BITMheader, sizeof(BITMheader));
			long dummy;
			file.Write(pOriginal->FindData("bits", B_RAW_TYPE, &dummy),
					   BITMheader.bits_length);
			file.Close();
			file.SetTypeAndApp('BITM', 'PPNT');
			
			// Create and return the B_SIMPLE_DATA message.
			BMessage *pMessage = new BMessage(B_SIMPLE_DATA);
			pMessage->AddRef("refs", file.Record()->Ref());
			return pMessage;
		}
	}
	
	// Oops, something went wrong.
	return 0;
}
