/*
 *     XaoS, a fast portable realtime fractal zoomer 
 *                  Copyright (C) 1996,1997 by
 *
 *      Jan Hubicka          (hubicka@paru.cas.cz)
 *      Thomas Marsh         (tmarsh@austin.ibm.com)
 *
 *	ui_be.cpp	BeOS user interface code, Jens Kilian (jjk@acm.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef __BEOS__

#include <stdio.h>
#include <Bitmap.h>
#include <InterfaceDefs.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>

#include "XaoSApplication.h"
#include "XaoSWindow.h"
#include "XaoSView.h"
#include "XaoSEvent.h"
#include "CLocker.h"

#include "ui.h"

static XaoSWindow *gpWindow;		// our BWindow
static XaoSView *gpView;			// our BView

static void *gpBuffers[2] = {		// fractal buffers
	0, 0							// (also display buffers in 8-bit mode)
};
static int gCurrentBuffer;
static BBitmap *gpTCBuffer;			// truecolor conversion buffer (32-bit only)

const int NCOLORS = 256;
static long gPalette[NCOLORS];		// truecolor palette
static int gNumColors = 0;

static thread_id main_thread;

struct ui_driver be_driver;
static int truecolor = 0;			// parameters settable from command line
static char *window_size = 0;
static int window_width = XSIZE;
static int window_height = YSIZE;

// Set up the user interface.
static int be_init(void)
{
	// Evaluate parameters.
	if (window_size) {
		int width, height;
		sscanf(window_size, "%dx%d", &width, &height);
		if (width > 0) {
			window_width = width;
		}
		if (height > 0) {
			window_height = height;
		}
	}

	if (truecolor) {
		be_driver.flags |= PALETTE_ROTATION|PALETTE_REDISPLAYS;
	}
	
	// Create the Application, Window and View.
	if (new XaoSApplication() != 0) {
		if ((gpView = new XaoSView(window_width, window_height)) != 0) {
			if ((gpWindow = new XaoSWindow(gpView)) != 0) {
				be_driver.textheight = gpView->TextHeight();
				
				// Spawn a thread to run the Application main loop in.
				// This is bass-ackwards from the normal way to do it,
				// but allows easier coexistence with the non-Be main().
				if ((main_thread = spawn_thread(XaoSApplication::Main,
												"XaoSApplication main loop",
												B_NORMAL_PRIORITY,
												be_app)) >= B_NO_ERROR) {
					resume_thread(main_thread);
					gpWindow->Show();
					return 1;
				}
				delete gpWindow;
			}
			delete gpView;
		}
		delete be_app;
	}
	
	main_thread = B_BAD_THREAD_ID;
	return 0;
}

// Return display buffer size.
static void be_get_size(int *x, int *y)
{
	*x = window_width;
	*y = window_height;
}

	
// XaoS has a somewhat different notion of buttons than BeOS ...
static int map_be_buttons(ulong beButtons, ulong modifiers)
{
	if ((modifiers & B_SHIFT_KEY) && (beButtons & B_PRIMARY_MOUSE_BUTTON)) {
		// Shift + left button -> middle, for DR8.2 (arrgh).
		return BUTTON2;
	}
	
	return ((beButtons & B_PRIMARY_MOUSE_BUTTON) ? BUTTON1 : 0)
		 | ((beButtons & B_TERTIARY_MOUSE_BUTTON) ? BUTTON2 : 0)
		 | ((beButtons & B_SECONDARY_MOUSE_BUTTON) ? BUTTON3 : 0);
}

// Get next event from keyboard, mouse, etc.
static inline bool check_key_state(key_info &info, int key_code)
{
	return info.key_states[key_code / 8] & (1 << (7 - key_code % 8));
}

static void be_processevents(int wait, int *x, int *y, int *buttons, int *k)
{
	static int currX = 0, currY = 0;
	static int currButtons = 0;
	static int cursorMask = 0;

	if (wait || port_count(gpView->EventPort()) > 0) {
		do {
			// Read events from our message port.
			long eventCode;
			XaoSEvent event;
			if (read_port(gpView->EventPort(), &eventCode, &event, sizeof(XaoSEvent))
				 < B_NO_ERROR) {
				 break;
			}
			
			// Handle event.
			switch (eventCode) {

			case XaoSEvent::Key:
				switch (event.keyEvent.key) {

				case B_LEFT_ARROW:
					cursorMask |= 1;
					break;
				
				case B_RIGHT_ARROW:
					cursorMask |= 2;
					break;
				
				case B_UP_ARROW:
					cursorMask |= 4;
					break;
				
				case B_DOWN_ARROW:
					cursorMask |= 8;
					break;

				default:
					if (event.keyEvent.key == '\033') {
						event.keyEvent.key = 'q';
					}
					if (ui_key(event.keyEvent.key) == 2) {
						return;
					}
					break;
				}
				break;

			case XaoSEvent::Mouse:
				currX = event.mouseEvent.x;
				currY = event.mouseEvent.y;
				currButtons = map_be_buttons(event.mouseEvent.buttons,
											 event.mouseEvent.modifiers);
				break;

			case XaoSEvent::Pulse:
				{	// Arrgh.  BeOS DR8.2 doesn't deliver key-up events.
					// We have to check the cursor keys periodically.
					key_info info;
					{	CLocker locker(*gpWindow);
						gpView->GetKeys(&info, FALSE);
					}
					if (!check_key_state(info, B_LEFT_ARROW)) {
						cursorMask &= ~1;
					}
					if (!check_key_state(info, B_RIGHT_ARROW)) {
						cursorMask &= ~2;
					}
					if (!check_key_state(info, B_UP_ARROW)) {
						cursorMask &= ~4;
					}
					if (!check_key_state(info, B_DOWN_ARROW)) {
						cursorMask &= ~8;
					}
				}
				break;

			case XaoSEvent::Resize:
				{	CLocker locker(*gpWindow);
					BRect bounds = gpView->Bounds();
					window_width = bounds.Width() + 1;
					window_height = bounds.Height() + 1;
				}
				ui_resize();
				break;
			}
			
		} while (port_count(gpView->EventPort()) > 0);
	}
	
	*x = currX;
	*y = currY;
	*buttons = currButtons;
	*k = cursorMask;
}

// Read the mouse.
static void be_getmouse(int *x, int *y, int *buttons)
{
	BPoint mouseLocation;
	ulong mouseButtons;
	{	CLocker locker(*gpWindow);
		gpView->GetMouse(&mouseLocation, &mouseButtons);
	}
	*x = (int)mouseLocation.x;
	*y = (int)mouseLocation.y;
	*buttons = map_be_buttons(mouseButtons, modifiers());
}

// Deallocate any stuff we allocated at startup.
static void be_uninitialise()
{
	if (main_thread != B_BAD_THREAD_ID) {
		// Tell the window that it may quit, and to do so.
		gpWindow->AllowQuit();
		gpWindow->PostMessage(B_QUIT_REQUESTED);
		
		// Wait until the application has shut down.
		long dummy;
		(void)wait_for_thread(main_thread, &dummy);
		delete be_app;
				
		main_thread = B_BAD_THREAD_ID;
	}
}

// Add color to palette.
// r,g,b are 0-255 range color channel values.
inline long rgb2bgra(int r, int g, int b)
{
	// Use BGRA order for fast conversion to BBitmap.
	return (b << 24) | (g << 16) | (r << 8);
}

static int be_set_color(int r, int g, int b, int init)
{
	if (!truecolor) {
		// BeOS 8-bit palette is fixed.
		return index_for_color(r, g, b);
		
	} else {
		if (init) {
			gNumColors = 0;
		} else if (gNumColors < NCOLORS) {
			gPalette[gNumColors] = rgb2bgra(r, g, b);
			return gNumColors++;
		}
	}
	return 0;
}

// Display c-string <text> at x,y
static void be_print(int x, int y, char *text)
{
	CLocker locker(*gpWindow);
	gpView->DrawText(BPoint(x, y), text);
}

// Display pixel buffer from fractal engine to window.
static void be_display(void)
{
	CLocker locker(*gpWindow);
	if (truecolor) {
		// Convert the current buffer to 32-bit format.
		unsigned char *src = (unsigned char *)(gpBuffers[gCurrentBuffer]) - 1;
		char *dst = (char *)(gpTCBuffer->Bits());
		long stride = gpTCBuffer->BytesPerRow();
		
		for (int y = window_height; y > 0; --y) {
			long *dp = (long *)dst - 1;
			for (int x = window_width; x > 0; --x) {
				*++dp = gPalette[*++src];
			}
			dst += stride;
		}
	}
 	gpView->Draw(gpView->Bounds());
}

// Allocate fractal engine pixel buffers.
static int be_alloc_buffers(char **buffer1, char **buffer2)
{
	if (!truecolor) {
		// In 8-bit mode, we let the fractal engine draw directly into one of
		// two BBitmaps, and flip between them.
		for (int i = 0; i < 2; ++i) {
			gpBuffers[i] =
				new BBitmap(BRect(B_ORIGIN,
								  BPoint(window_width-1, window_height-1)),
							B_COLOR_8_BIT);
		}
		gCurrentBuffer = 0;
		
		if (gpBuffers[0] && gpBuffers[1]) {
			*buffer1 = (char *)(((BBitmap *)(gpBuffers[0]))->Bits());
			*buffer2 = (char *)(((BBitmap *)(gpBuffers[1]))->Bits());
		
			CLocker locker(*gpWindow);
			gpView->SetBuffer((BBitmap *)(gpBuffers[0]));
			
			return ((BBitmap *)(gpBuffers[0]))->BytesPerRow();
		}

	} else {
		// In true-color mode, we let the fractal engine draw into one of two buffers
		// and convert to a BBitmap before drawing.
		long size = window_width * window_height;
		for (int i = 0; i < 2; ++i) {
			gpBuffers[i] = new unsigned char[size];
		}
		gCurrentBuffer = 0;
		gpTCBuffer = new BBitmap(BRect(B_ORIGIN,
									   BPoint(window_width-1, window_height-1)),
								 B_RGB_32_BIT);

		if (gpBuffers[0] && gpBuffers[1] && gpTCBuffer) {
			*buffer1 = (char *)(gpBuffers[0]);
			*buffer2 = (char *)(gpBuffers[1]);
		
			CLocker locker(*gpWindow);
			gpView->SetBuffer(gpTCBuffer);
			
			return window_width;
		}
	}

	return 0;
}

// Deallocate fractal engine pixel buffers.
static void be_free_buffers(char * /*buffer1*/, char * /*buffer2*/)
{
	CLocker locker(*gpWindow);
	gpView->SetBuffer(0);

	for (int i = 0; i < 2; ++i) {
		if (!truecolor) {
			delete (BBitmap *)(gpBuffers[i]);
		} else {
			delete [] (unsigned char *)(gpBuffers[i]);
		}
		gpBuffers[i] = 0;
	}
}

// Switch current pixel buffer to copy from on next display update.
static void be_flip_buffers(void)
{
	gCurrentBuffer ^= 1;
	if (!truecolor) {
		CLocker locker(*gpWindow);
		gpView->SetBuffer((BBitmap *)(gpBuffers[gCurrentBuffer]));
	}
}

// Rotate the palette (only available in 32-bit mode, otherwise palette is fixed).
static void be_rotate_palette(int /*direction*/)
{
	zoom_context *c = ui_getcontext();
	int ncolors = c->num_colors;
	for (int i = 0; i < ncolors; ++i) {
		gPalette[i] = rgb2bgra(c->cmap[0][i], c->cmap[1][i], c->cmap[2][i]);
	}
	be_display();
}

static char *helptext[] =
{
    "BeOS DRIVER VERSION 1.1                ",
    "=======================================",
    "Written by Jens Kilian <jjk@acm.org>   ",
    "                                       ",
    "Features:                              ",
    "                                       ",
    "- Use double-click to drag the current ",
    "  image to another application (e.g.,  ",
    "  ImageViewer or the Browser).         ",
    "                                       ",
    "Restrictions:                          ",
    "                                       ",
    "- Keyboard and mouse handling are      ",
    "  a bit quirky thanks to the less than ",
    "  truthful event reporting in DR8.2.   ",
    "  Use SHIFT + left mouse button to pan.",
    "                                       ",
    "- Palette rotation is only available   ",
    "  in 32-bit (-truecolor) mode.         ",
    "                                       ",
    "- Screen size computations assume a 17\"",
    "  monitor at resolution 1024x768.      ",
    "                                       ",
    "- The window cannot be resized because ",
    "  XaoS (currently :-) doesn't support  ",
    "  live resizing.                       ",
    "                                       ",
    "- The icon truly sucks. (Hint, hint.)  ",
    "                                       ",
    "  Remember - don't dream it, Be it.    ",
    "                                       "
};

#define UGLYTEXTSIZE (sizeof(helptext)/sizeof(char *))

static struct params params[] =
{
    { "-truecolor", P_SWITCH, &truecolor, "true color (32 bit) mode" },
    { "-size", P_STRING, &window_size, "set window size (WIDTHxHEIGHT)" },
    { NULL, 0, NULL, NULL }
};

struct ui_driver be_driver =
{
    "BeOS",
    be_init,
    be_get_size,
    be_processevents,
    be_getmouse,
    be_uninitialise,
    be_set_color,
    be_print,
    be_display,
    be_alloc_buffers,
    be_free_buffers,
    be_flip_buffers,
    be_rotate_palette,
    NCOLORS,			// max palette size
    0,					// text height (see be_init)
    helptext,			// help text, size
    UGLYTEXTSIZE,
    params,				// command-line parameters
    UPDATE_AFTER_PALETTE|PIXELSIZE|RESOLUTION,
    0.02979, 0.02930,	// this is for my monitor in 1024x768
    1024, 768
};
#endif	// __BEOS__